@@ -4,7 +4,7 @@
 # may not use this file except in compliance with the License.  You
 # may obtain a copy of the License at
 #
-# http://www.apache.org/licenses/LICENSE-2.0
+# https://www.apache.org/licenses/LICENSE-2.0
 #
 # Unless required by applicable law or agreed to in writing, software
 # distributed under the License is distributed on an "AS IS" BASIS,
@@ -16,6 +16,7 @@
 
 from __future__ import annotations
 
+import asyncio
 import logging
 import os
 import queue
@@ -29,7 +30,7 @@ from typing import TYPE_CHECKING, Any, Callable, Mapping, Optional, cast
 
 from pymongo import _csot, common, helpers_shared, periodic_executor
 from pymongo.asynchronous.client_session import _ServerSession, _ServerSessionPool
-from pymongo.asynchronous.monitor import SrvMonitor
+from pymongo.asynchronous.monitor import MonitorBase, SrvMonitor
 from pymongo.asynchronous.pool import Pool
 from pymongo.asynchronous.server import Server
 from pymongo.errors import (
@@ -40,6 +41,7 @@ from pymongo.errors import (
     OperationFailure,
     PyMongoError,
     ServerSelectionTimeoutError,
+    WaitQueueTimeoutError,
     WriteError,
 )
 from pymongo.hello import Hello
@@ -118,8 +120,8 @@ class Topology:
         if _SDAM_LOGGER.isEnabledFor(logging.DEBUG):
             _debug_log(
                 _SDAM_LOGGER,
-                topologyId=self._topology_id,
                 message=_SDAMStatusMessage.START_TOPOLOGY,
+                topologyId=self._topology_id,
             )
 
         if self._publish_tp:
@@ -150,10 +152,10 @@ class Topology:
         if _SDAM_LOGGER.isEnabledFor(logging.DEBUG):
             _debug_log(
                 _SDAM_LOGGER,
+                message=_SDAMStatusMessage.TOPOLOGY_CHANGE,
                 topologyId=self._topology_id,
                 previousDescription=initial_td,
                 newDescription=self._description,
-                message=_SDAMStatusMessage.TOPOLOGY_CHANGE,
             )
 
         for seed in topology_settings.seeds:
@@ -163,10 +165,10 @@ class Topology:
             if _SDAM_LOGGER.isEnabledFor(logging.DEBUG):
                 _debug_log(
                     _SDAM_LOGGER,
+                    message=_SDAMStatusMessage.START_SERVER,
                     topologyId=self._topology_id,
                     serverHost=seed[0],
                     serverPort=seed[1],
-                    message=_SDAMStatusMessage.START_SERVER,
                 )
 
         # Store the seed list to help diagnose errors in _error_message().
@@ -207,6 +209,9 @@ class Topology:
         if self._settings.fqdn is not None and not self._settings.load_balanced:
             self._srv_monitor = SrvMonitor(self, self._settings)
 
+        # Stores all monitor tasks that need to be joined on close or server selection
+        self._monitor_tasks: list[MonitorBase] = []
+
     async def open(self) -> None:
         """Start monitoring, or restart after a fork.
 
@@ -232,9 +237,7 @@ class Topology:
             warnings.warn(  # type: ignore[call-overload] # noqa: B028
                 "AsyncMongoClient opened before fork. May not be entirely fork-safe, "
                 "proceed with caution. See PyMongo's documentation for details: "
-                "https://www.mongodb.com/docs/languages/"
-                "python/pymongo-driver/current/faq/"
-                "#is-pymongo-fork-safe-",
+                "https://dochub.mongodb.org/core/pymongo-fork-deadlock",
                 **kwargs,
             )
             async with self._lock:
@@ -283,6 +286,10 @@ class Topology:
         else:
             server_timeout = server_selection_timeout
 
+        # Cleanup any completed monitor tasks safely
+        if not _IS_SYNC and self._monitor_tasks:
+            await self.cleanup_monitors()
+
         async with self._lock:
             server_descriptions = await self._select_servers_loop(
                 selector, server_timeout, operation, operation_id, address
@@ -347,7 +354,7 @@ class Topology:
                     operationId=operation_id,
                     topologyDescription=self.description,
                     clientId=self.description._topology_settings._topology_id,
-                    remainingTimeMS=int(end_time - time.monotonic()),
+                    remainingTimeMS=int(1000 * (end_time - time.monotonic())),
                 )
                 logged_waiting = True
 
@@ -493,7 +500,6 @@ class Topology:
 
         self._description = new_td
         await self._update_servers()
-        self._receive_cluster_time_no_lock(server_description.cluster_time)
 
         if self._publish_tp and not suppress_event:
             assert self._events is not None
@@ -506,10 +512,10 @@ class Topology:
         if _SDAM_LOGGER.isEnabledFor(logging.DEBUG) and not suppress_event:
             _debug_log(
                 _SDAM_LOGGER,
+                message=_SDAMStatusMessage.TOPOLOGY_CHANGE,
                 topologyId=self._topology_id,
                 previousDescription=td_old,
                 newDescription=self._description,
-                message=_SDAMStatusMessage.TOPOLOGY_CHANGE,
             )
 
         # Shutdown SRV polling for unsupported cluster types.
@@ -520,12 +526,8 @@ class Topology:
             and self._description.topology_type not in SRV_POLLING_TOPOLOGIES
         ):
             await self._srv_monitor.close()
-
-        # Clear the pool from a failed heartbeat.
-        if reset_pool:
-            server = self._servers.get(server_description.address)
-            if server:
-                await server.pool.reset(interrupt_connections=interrupt_connections)
+            if not _IS_SYNC:
+                self._monitor_tasks.append(self._srv_monitor)
 
         # Wake anything waiting in select_servers().
         self._condition.notify_all()
@@ -549,6 +551,11 @@ class Topology:
             # that didn't include this server.
             if self._opened and self._description.has_server(server_description.address):
                 await self._process_change(server_description, reset_pool, interrupt_connections)
+        # Clear the pool from a failed heartbeat, done outside the lock to avoid blocking on connection close.
+        if reset_pool:
+            server = self._servers.get(server_description.address)
+            if server:
+                await server.pool.reset(interrupt_connections=interrupt_connections)
 
     async def _process_srv_update(self, seedlist: list[tuple[str, Any]]) -> None:
         """Process a new seedlist on an opened topology.
@@ -572,10 +579,10 @@ class Topology:
         if _SDAM_LOGGER.isEnabledFor(logging.DEBUG):
             _debug_log(
                 _SDAM_LOGGER,
+                message=_SDAMStatusMessage.TOPOLOGY_CHANGE,
                 topologyId=self._topology_id,
                 previousDescription=td_old,
                 newDescription=self._description,
-                message=_SDAMStatusMessage.TOPOLOGY_CHANGE,
             )
 
     async def on_srv_update(self, seedlist: list[tuple[str, Any]]) -> None:
@@ -695,6 +702,8 @@ class Topology:
             old_td = self._description
             for server in self._servers.values():
                 await server.close()
+                if not _IS_SYNC:
+                    self._monitor_tasks.append(server._monitor)
 
             # Mark all servers Unknown.
             self._description = self._description.reset()
@@ -705,6 +714,8 @@ class Topology:
             # Stop SRV polling thread.
             if self._srv_monitor:
                 await self._srv_monitor.close()
+                if not _IS_SYNC:
+                    self._monitor_tasks.append(self._srv_monitor)
 
             self._opened = False
             self._closed = True
@@ -734,13 +745,13 @@ class Topology:
         if _SDAM_LOGGER.isEnabledFor(logging.DEBUG):
             _debug_log(
                 _SDAM_LOGGER,
+                message=_SDAMStatusMessage.TOPOLOGY_CHANGE,
                 topologyId=self._topology_id,
                 previousDescription=old_td,
                 newDescription=self._description,
-                message=_SDAMStatusMessage.TOPOLOGY_CHANGE,
             )
             _debug_log(
-                _SDAM_LOGGER, topologyId=self._topology_id, message=_SDAMStatusMessage.STOP_TOPOLOGY
+                _SDAM_LOGGER, message=_SDAMStatusMessage.STOP_TOPOLOGY, topologyId=self._topology_id
             )
 
         if self._publish_server or self._publish_tp:
@@ -879,6 +890,8 @@ class Topology:
                 # Clear the pool.
                 await server.reset(service_id)
         elif isinstance(error, ConnectionFailure):
+            if isinstance(error, WaitQueueTimeoutError):
+                return
             # "Client MUST replace the server's description with type Unknown
             # ... MUST NOT request an immediate check of the server."
             if not self._settings.load_balanced:
@@ -944,6 +957,8 @@ class Topology:
         for address, server in list(self._servers.items()):
             if not self._description.has_server(address):
                 await server.close()
+                if not _IS_SYNC:
+                    self._monitor_tasks.append(server._monitor)
                 self._servers.pop(address)
 
     def _create_pool_for_server(self, address: _Address) -> Pool:
@@ -1031,6 +1046,15 @@ class Topology:
             else:
                 return ",".join(str(server.error) for server in servers if server.error)
 
+    async def cleanup_monitors(self) -> None:
+        tasks = []
+        try:
+            while self._monitor_tasks:
+                tasks.append(self._monitor_tasks.pop())
+        except IndexError:
+            pass
+        await asyncio.gather(*[t.join() for t in tasks], return_exceptions=True)  # type: ignore[func-returns-value]
+
     def __repr__(self) -> str:
         msg = ""
         if not self._opened:
